from typing import List, Set
from dataclasses import dataclass, field
import sys


@dataclass
class GeneStats:
    """Statistics for a gene detection.

    Only sequences meeting min_identity threshold are counted.
    - gene_coverage: Percentage of the gene/subject sequence covered by all alignments combined
    - avg_identity: Average identity across all qualifying sequences for this gene
    """
    gene_name: str
    gene_length: int = 0  # Length of the gene in the database
    num_sequences: int = 0  # Number of sequences passing min_identity
    gene_coverage: float = 0.0  # Percentage of gene covered by all alignments
    avg_identity: float = 0.0
    identities: List[float] = field(default_factory=list)
    covered_positions: Set[int] = field(default_factory=set)  # All positions covered on the gene

    def add_hit(self, sstart: int, send: int, identity: float, gene_len: int = 0):
        """Add a hit to the statistics (only called for sequences passing min_identity).

        Args:
            sstart: Start position on subject/gene (1-based, inclusive)
            send: End position on subject/gene (1-based, inclusive)
            identity: Percent identity of this alignment
            gene_len: Length of the gene (if known)
        """
        self.num_sequences += 1
        self.identities.append(identity)

        # Add all positions covered by this alignment (convert to 0-based for consistency)
        start = min(sstart, send) - 1  # Convert to 0-based
        end = max(sstart, send)  # Inclusive end in 1-based becomes exclusive in 0-based
        for pos in range(start, end):
            self.covered_positions.add(pos)

        if gene_len > 0:
            self.gene_length = max(self.gene_length, gene_len)

    def add_positions(self, positions: Set[int], identity: float, gene_len: int = 0):
        """Add a set of positions directly (for BAM parsing where we track exact positions).

        Args:
            positions: Set of 0-based positions covered on the gene
            identity: Percent identity of this alignment
            gene_len: Length of the gene (if known)
        """
        self.num_sequences += 1
        self.identities.append(identity)
        self.covered_positions.update(positions)

        if gene_len > 0:
            self.gene_length = max(self.gene_length, gene_len)

    def finalise(self):
        """Calculate final statistics."""
        if self.num_sequences > 0:
            self.avg_identity = sum(self.identities) / self.num_sequences

        # Calculate gene coverage as percentage of gene length covered
        if self.gene_length > 0:
            self.gene_coverage = (len(self.covered_positions) / self.gene_length) * 100


# @dataclass
# class GeneStats:
#     """Statistics for a gene detection.
#     Only sequences meeting min_identity threshold are counted.
#     - gene_coverage: Percentage of the gene/subject sequence covered by all alignments combined
#     - avg_identity: Average identity across all qualifying sequences for this gene
#     """
#     gene_name: str
#     gene_length: int = 0  # Length of the gene in the database
#     num_sequences: int = 0  # Number of sequences passing min_identity
#     gene_coverage: float = 0.0  # Percentage of gene covered by all alignments
#     avg_identity: float = 0.0
#     identities: List[float] = field(default_factory=list)
#     covered_positions: Set[int] = field(default_factory=set)  # All positions covered on the gene
#
#     def add_hit(self, sstart: int, send: int, identity: float, gene_len: int = 0):
#         """Add a hit to the statistics (only called for sequences passing min_identity).
#         Args:
#             sstart: Start position on subject/gene
#             send: End position on subject/gene
#             identity: Percent identity of this alignment
#             gene_len: Length of the gene (if known)
#         """
#         self.num_sequences += 1
#         self.identities.append(identity)
#
#         # Add all positions covered by this alignment
#         start = min(sstart, send)
#         end = max(sstart, send)
#         for pos in range(start, end + 1):
#             self.covered_positions.add(pos)
#
#         if gene_len > 0:
#             self.gene_length = max(self.gene_length, gene_len)
#         else:
#             sys.exit("gene length not known")
#
#     def finalise(self):
#         """Calculate final statistics."""
#         if self.num_sequences > 0:
#             self.avg_identity = sum(self.identities) / self.num_sequences
#
#         # Calculate gene coverage as percentage of gene length covered
#         if self.gene_length > 0:
#             self.gene_coverage = (len(self.covered_positions) / self.gene_length) * 100
#



import gzip
import os
import sys
import logging

logging.basicConfig(level=logging.INFO, stream=sys.stdout,
                    format='%(asctime)s %(levelname)s %(name)s: %(message)s')
logger = logging.getLogger(__name__)

def FASTQ_to_FASTA(options):
    # If Paired-FASTQ, convert R1/R2 FASTQ -> FASTA and set options.input to the combined FASTA
    logger.info("FASTQ_to_FASTA: using seqtk to convert paired FASTQ -> combined FASTA")
    import subprocess
    import shutil

    def find_pair(input_spec):
        if ',' in input_spec:
            r1, r2 = map(str.strip, input_spec.split(',', 1))
            return r1, r2
        base = input_spec
        candidates = [base, base + '_R1.fastq', base + '_R1.fq', base + '_1.fastq', base + '_1.fq']
        r1_path = None
        for c in candidates:
            if os.path.exists(c):
                r1_path = c
                break
        if not r1_path:
            logger.error("Could not locate R1 FASTQ. Provide `R1.fastq,R2.fastq` as `-i`.")
            sys.exit(1)
        if '_R1.' in r1_path:
            r2_path = r1_path.replace('_R1.', '_R2.')
        elif '_1.' in r1_path:
            r2_path = r1_path.replace('_1.', '_2.')
        else:
            r2_path = r1_path.replace('_R1', '_R2')
        return r1_path, r2_path

    r1_path, r2_path = find_pair(options.input)
    if not os.path.exists(r1_path) or not os.path.exists(r2_path):
        logger.error(f"Paired FASTQ files not found: {r1_path}, {r2_path}")
        sys.exit(1)

    conv_dir = os.path.join(options.output, 'paired_fastq_fasta')
    os.makedirs(conv_dir, exist_ok=True)
    combined_fasta = os.path.join(conv_dir, 'fastq_to_fasta_combined.fasta.gz')
    if os.path.exists(combined_fasta):
        logger.info(f"Found existing combined FASTA at `{combined_fasta}`; using it (skipping conversion)")
        options.fasta_input = combined_fasta
        options.fastq_input = (r1_path, r2_path)
        return

    # ensure seqtk is available
    if shutil.which('seqtk') is None:
        logger.error("`seqtk` not found in PATH. Install seqtk or provide a FASTA input.")
        sys.exit(1)

    def seqtk_fastq_to_fasta_stream(fastq_path, out_handle):
        cmd = ['seqtk', 'seq', '-A', fastq_path]
        logger.info(f"Running: {' '.join(cmd)}")
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if proc.stdout is None:
            logger.error("Failed to start seqtk process")
            sys.exit(1)
        try:
            for chunk in iter(lambda: proc.stdout.read(8192), b''):
                if not chunk:
                    break
                out_handle.write(chunk)
        finally:
            proc.stdout.close()
            _, stderr = proc.communicate()
            if proc.returncode != 0:
                logger.error(f"seqtk failed: {stderr.decode(errors='ignore')}")
                sys.exit(1)

    # convert both FASTQ files and append into a single gzipped FASTA
    with gzip.open(combined_fasta, 'wb') as out:
        seqtk_fastq_to_fasta_stream(r1_path, out)
        seqtk_fastq_to_fasta_stream(r2_path, out)

    logger.info(f"Combined FASTA created at {combined_fasta}")
    options.fasta_input = combined_fasta
    options.fastq_input = (r1_path, r2_path)


import argparse
import sys
import os

try:
    from .constants import *
    from .databases import RESFINDER_DATABASES, CARD_DATABASES
    from .workflow import AMRWorkflow
    from .gene_stats import GeneStats
    from .utils import *
except (ModuleNotFoundError, ImportError, NameError, TypeError) as error:
    from constants import *
    from databases import RESFINDER_DATABASES, CARD_DATABASES
    from workflow import AMRWorkflow
    from gene_stats import GeneStats
    from utils import *

def main():
    parser = argparse.ArgumentParser(
        description='AMRfíor ' + AMRFIOR_VERSION + ' - The Multi-Tool AMR Gene Detection Workflow.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Basic usage with default tools (runs DNA & protein tools)
  AMRfior -i reads.fasta -st Single-FASTA -o results/

  # Select specific tools and output detected FASTA sequences
  AMRfior -i reads.fasta -st Single-FASTA -o results/ \
    --tools diamond bowtie2 \
    --report_fasta detected

  # Custom thresholds, paired-fastq input, threads and dna-only mode
  AMRfior -i reads_R1.fastq,reads_R2.fastq -st Paired-FASTQ -o results/ \
    -t 16 --d-min-cov 90 --d-min-id 85 \
    --dna-only
        """
    )

    # Required arguments
    required_group = parser.add_argument_group('Required selection')
    required_group.add_argument('-i', '--input', required=True,
                        help='Input FASTA/FASTAQ file(s) with sequences to analyse - Separate FASTQ R1 and R2 '
                             'with a comma for Paired-FASTQ or single file path for Single-FASTA - .gz files accepted')
    required_group.add_argument('-st', '--sequence-type', required=True,
                        choices=['Single-FASTA', 'Paired-FASTQ'],
                        help='Specify the input Sequence Type: Single-FASTA or Paired-FASTQ (R1+R2) - Will '
                             'convert Paired-FASTQ to single combined FASTA for BLAST and DIAMOND analyses (SLOW)')
    required_group.add_argument('-o', '--output', required=True,
                        help='Output directory for results')

    # Output selection
    output_group = parser.add_argument_group('Output selection')
    output_group.add_argument('--report_fasta',
                            choices=['None', 'all', 'detected', 'detected-all'], #, 'hmmer_dna', 'hmmer_protein'],
                            default=[None], #, 'hmmer_dna','hmmer_protein'],
                            help='Specify whether to output sequences that "mapped" to genes.'
                                 '"all" should only be used for deep investigation/debugging.'
                                 '"detected" will report the reads that passed detection thresholds for each detected gene.'
                                 '"detected-all" will report all reads for each detected gene.  (default: None)')

    # Tool selection
    tool_group = parser.add_argument_group('Tool selection')
    tool_group.add_argument('--tools', nargs='+',
                            choices=['blastn', 'blastx', 'diamond', 'bowtie2', 'bwa', 'minimap2', 'all'], #, 'hmmer_dna', 'hmmer_protein'],
                            default=['diamond', 'bowtie2', 'bwa', 'minimap2'], #, 'hmmer_dna','hmmer_protein'],
                            help='Specify which tools to run - "all" will run all tools'
                                 ' (default: all except blastx/n as it is very slow!!)')

    query_threshold_group = parser.add_argument_group('Query threshold Parameters')
    query_threshold_group.add_argument('--q-min-cov', '--query-min-coverage', type=float, default=40.0,
                                      dest='query_min_coverage',
                                      help='Minimum coverage threshold in percent (default: 40.0)')

    gene_detection_group = parser.add_argument_group('Gene Detection Parameters')
    gene_detection_group.add_argument('--d-min-cov', '--detection-min-coverage', type=float, default=80.0,
                              dest='detection_min_coverage',
                              help='Minimum coverage threshold in percent (default: 80.0)')
    gene_detection_group.add_argument('--d-min-id', '--detection-min-identity', type=float, default=80.0,
                              dest='detection_min_identity',
                              help='Minimum identity threshold in percent (default: 80.0)')
    # gene_detection_group.add_argument( '--max_target_seqs', dest='max_target_seqs', type=int, default=100,
    #                           help='Maximum number of "hits" to return per query sequence (default: 100)')


    # Mode selection
    mode_group = parser.add_argument_group('Mode Selection')
    mode_group.add_argument('--dna-only', action='store_true',
                            help='Run only DNA-based tools')
    mode_group.add_argument('--protein-only', action='store_true',
                            help='Run only protein-based tools')
    mode_group.add_argument('--sensitivity', type=str, default='default',
                            choices=['default', 'conservative', 'sensitive', 'very-sensitive'],
                            help='Preset sensitivity levels - default means each tool uses its own default settings and '
                                 'very-sensitive applies DIAMONDs --ultra-sensitive and Bowtie2s'
                                 ' --very-sensitive-local presets')

    # Tool-specific parameters
    tool_params_group = parser.add_argument_group('Tool-Specific Parameters')
    tool_params_group.add_argument('--minimap2-preset', default='sr',
                                   choices=['sr', 'map-ont', 'map-pb', 'map-hifi'],
                                   help='Minimap2 preset: sr=short reads, map-ont=Oxford Nanopore, '
                                        'map-pb=PacBio, map-hifi=PacBio HiFi (default: sr)')
    # tool_params_group.add_argument('-e', '--evalue', type=float, default=1e-10,
    #                                help='E-value threshold (default: 1e-10)')

    # Runtime parameters
    runtime_group = parser.add_argument_group('Runtime Parameters')
    runtime_group.add_argument('-t', '--threads', type=int, default=4,
                              help='Number of threads to use (default: 4)')
    runtime_group.add_argument('--no_cleanup',  action='store_true',)
    runtime_group.add_argument('-v', '--verbose', action='store_true',)

    options = parser.parse_args()

    if options.report_fasta == ['None']:
        options.report_fasta = None

    # FASTA/FASTQ handling
    if options.sequence_type == 'Paired-FASTQ':
        if any(tool in ('blastn', 'blastx', 'diamond', 'all') for tool in (options.tools or [])):
            FASTQ_to_FASTA(options)
        else:
            options.fastq_input = options.input
            options.fasta_input = None
    else:
        # Check input file exists
        if not os.path.exists(options.input):
            print(f"Error: Input file '{options.input}' not found", file=sys.stderr)
            sys.exit(1)
        options.fasta_input = options.input
        options.fastq_input = None

    if options.tools == ['all']:
        options.tools = ['blastn', 'blastx', 'diamond', 'bowtie2', 'bwa', 'minimap2']  #, 'hmmer_dna','hmmer_protein']

    # Load database paths from databases.py
    resfinder_dbs = {tool: RESFINDER_DATABASES.get(tool) for tool in options.tools if RESFINDER_DATABASES.get(tool)}
    card_dbs = {tool: CARD_DATABASES.get(tool) for tool in options.tools if CARD_DATABASES.get(tool)}

    if not resfinder_dbs and not card_dbs:
        print("Error: At least one database must be specified in databases.py", file=sys.stderr)
        sys.exit(1)

    # Determine run modes
    run_dna = True
    run_protein = True

    if options.dna_only:
        run_protein = False
    if options.protein_only:
        run_dna = False

    if not run_dna and not run_protein:
        print("Error: Cannot disable both DNA and protein modes", file=sys.stderr)
        sys.exit(1)



    # Tool sensitivity
    tool_sensitivity_params = {}

    if hasattr(options, 'sensitivity') and options.sensitivity == 'default':
        # Use each tool's default sensitivity settings
        pass
    elif hasattr(options, 'sensitivity') and options.sensitivity == 'very-sensitive':
        # Example: set sensitivity for supported tools
        tool_sensitivity_params['bowtie2'] = {'sensitivity': '--very-sensitive-local'}
        tool_sensitivity_params['diamond'] = {'sensitivity': '--ultra-sensitive'}

    # Run Workflow
    workflow = AMRWorkflow(
        input_fasta=options.fasta_input,
        input_fastq=options.fastq_input,
        output_dir=options.output,
        resfinder_dbs=resfinder_dbs,
        card_dbs=card_dbs,
        threads=options.threads,
        tool_sensitivity_params=tool_sensitivity_params,
        #max_target_seqs=options.max_target_seqs,
        #evalue=options.evalue,
        detection_min_coverage=options.detection_min_coverage,
        detection_min_identity=options.detection_min_identity,
        query_min_coverage=options.query_min_coverage,
        run_dna=run_dna,
        run_protein=run_protein,
        sequence_type=options.sequence_type,
        report_fasta=options.report_fasta,
        no_cleanup=options.no_cleanup,
        verbose=options.verbose
    )

    results = workflow.run_workflow(options)

    # Exit with error code if all tools failed
    all_failed = True
    for db_results in results.values():
        for success, _ in db_results.values():
            if success:
                all_failed = False
                break
        if not all_failed:
            break

    if all_failed:
        sys.exit(1)


if __name__ == "__main__":
    main()